
class BaseCommand(object):
    """Interface for package commands"""

    name = None

    def register_argparser(self, parent_parser):
        """Register command line parser"""
        raise NotImplementedError

    def do(self, args, config):
        """Do the task that this command needs to do"""
        raise NotImplementedError


class CloneCommand(BaseCommand):
    """Clone a package repository from dist-git"""

    name = 'clone'

    def register_argparser(self, parent_subparser):
        clone_parser = parent_subparser.add_parser(
            self.name,
            help='Clone and checkout a module',
            description='This command will clone the named module from the '
                        'configured repository base URL. By default it will '
                        'also checkout the master branch for your working '
                        'copy.')
        self.parser = clone_parser

        # Allow an old style clone with subdirs for branches
        clone_parser.add_argument(
            '--branches', '-B',
            action='store_true',
            help='Do an old style checkout with subdirs for branches')
        # provide a convenient way to get to a specific branch
        clone_parser.add_argument(
            '--branch', '-b',
            help='Check out a specific branch')
        # allow to clone without needing a account on the scm server
        clone_parser.add_argument(
            '--anonymous', '-a',
            action='store_true',
            help='Check out a module anonymously')
        # store the module to be cloned
        clone_parser.add_argument(
            'module',
            nargs=1,
            help='Name of the module to clone')
        # Eventually specify where to clone the module
        clone_parser.add_argument(
            "clone_target",
            default=None,
            nargs="?",
            help='Directory in which to clone the module')
        clone_parser.set_defaults(command=self.clone)

        # Add an alias for historical reasons
        co_parser = self.subparsers.add_parser(
            'co',
            parents=[clone_parser],
            conflict_handler='resolve',
            help='Alias for clone')
        co_parser.set_defaults(command=self.clone)

    def do(self, args, config):
        print 'Run:', 'git clone', args.package


class DiffCommand(BaseCommand):
    """Show diff in this repository"""

    name = 'diff'

    def register_argparser(self, parent_subparser):
        self.parser = parent_subparser.add_parser(self.name)

    def do(self, args, config):
        print 'Run', 'git diff'


class SourcesCommand(BaseCommand):
    """Download sources files"""

    name = 'sources'

    def register_argparser(self, parent_subparser):
        self.parser = parent_subparser.add_parser(
            self.name,
            description='Download sources files.')

    def do(self, args, config):
        print 'sources: download sources files'


class NewSourcesCommand(BaseCommand):
    """Upload one or more sources files"""

    name = 'new-sources'

    def register_argparser(self, parent_subparser):
        self.parser = parent_subparser.add_parser(self.name)

    def do(self, args, config):
        print 'upload sources file:', args, sources_files


class PrepCommand(BaseCommand):
    """Run rpmbuild prep"""

    name = 'prep'

    def register_argparser(self, parent_subparser):
        self.parser = parent_subparser.add_parser(self.name)

    def do(self, args, config):
        print 'Run rpmbuild prep'

